Stream = require('stream')
Zlib = require('zlib')

Request = require('request')
SocketIoStream = require('socket.io-stream')

Base = require('../base')
Logger = require('./logger')


debug = Base.logger('http_proxy')

PROXY_HTTP_HOST_USE_PROXY = 'USE_PROXY'

upstreamUrl = null
isUpstreamHttps = null

# Rewriting vars.
isRewriteHostOn = null
upstreamHost = null
upstreamProto = null
upstreamHostRe = null
rewriteProtoDelimReStr = ":\\\\?\\/\\\\?\\/" # supports '://'  or ':\/\/'
rewriteCustomReplacers = null


exports.setupHttpProxy = (upstream) ->
  upstreamUrl = upstream || Base.env.IX_HTTP_PROXY
  upstreamUrl = upstreamUrl?.trim().replace(/\/$/, '')
  throw new Error("No upstream server defined.") unless upstreamUrl
  debug('setupHttpProxy upstreamUrl', upstreamUrl)
  isUpstreamHttps = upstreamUrl.slice(0,6) == 'https:'
  isRewriteHostOn = Base.env.IX_PROXY_REWRITE_HOST?.toLowerCase() == 'on'
  if isRewriteHostOn
    upstreamHost = upstreamUrl.replace(/^https?:\/\//i,'')
    upstreamProto = if isUpstreamHttps then 'https' else 'http'
    upstreamHostReStr = regexEscape(upstreamHost)
    upstreamHostRe = new RegExp("\\b#{upstreamHostReStr}\\b", "ig")
    if Base.env.IX_PROXY_PROTO_DELIM_REGEX
      rewriteProtoDelimReStr += "|"+Base.env.IX_PROXY_PROTO_DELIM_REGEX
    customRes = Base.env.IX_PROXY_REWRITE_CUSTOM_REGEXES
    if customRes
      console.log("Parsing IX_PROXY_REWRITE_CUSTOM_REGEXES: #{customRes}")
      replacers = JSON.parse(customRes)
      rewriteCustomReplacers = []
      for replacer in replacers
        flags = "g"
        flags += replacer[2] if replacer.length > 2
        rewriteCustomReplacers.push([new RegExp(replacer[0], flags), replacer[1]])
      debug('rewriteCustomReplacers', rewriteCustomReplacers)

exports.onHttpProxy = (faasReq, raw_cb_proxyResponse) ->
  debug('onHttpProxy.faasReq.headers', faasReq?.headers)
  faasReq.origHost = faasReq.headers['x-forwarded-host']
  faasReq.origProto = faasReq.headers['x-forwarded-proto']
  Logger.logHttpRequest(faasReq)
  cb_proxyResponse = (pr) ->
    debug('cb_proxyResponse', { headers: pr?.headers, lag: pr?.lag, statusCode: pr?.statusCode, statusMessage: pr?.statusMessage, error: pr?.error })
    # debug('cp_proxyResponse.responseStream', pr?.responseStream)
    raw_cb_proxyResponse(pr)
  url = upstreamUrl + faasReq.url
  fwdOptions =
    url: url
    headers: faasReq.headers
    method: faasReq.method
    followRedirect: false
  httpHost = Base.env.IX_PROXY_HTTP_HOST
  if isUpstreamHttps || httpHost == PROXY_HTTP_HOST_USE_PROXY
    delete fwdOptions.headers.host
  if httpHost && httpHost != PROXY_HTTP_HOST_USE_PROXY
    fwdOptions.headers.host = httpHost
  if isRewriteHostOn
    delete fwdOptions.headers.host
    delete fwdOptions.headers['x-forwarded-host']
    delete fwdOptions.headers['x-forwarded-server']
  debug('request', fwdOptions)
  fwdReq = null
  requestStream = faasReq.requestStream
  proxyResponse = {}
  reqCounter = new Stream.Transform()
  reqCounter.numBytes = 0
  reqCounter._transform = (chunk, encoding, callback) ->
    debug('request chunk', encoding, chunk)
    console.log(chunk.toString())
    reqCounter.numBytes += chunk.length
    callback(null, chunk)
  try
    fwdReq = Request(fwdOptions)
  catch e
    # Error: Invalid protocol: localhost:
    proxyResponse.statusCode = 500
    proxyResponse.error = e.toString()
    Logger.logHttpResponseError(faasReq, proxyResponse)
    cb_proxyResponse(proxyResponse)
    requestStream.on('data', (data) -> debug('fwdReq.error data', data) )
    # requestStream.end(e.toString())
    return
  responseStarted = false
  responseStream = SocketIoStream.createStream()
  proxyResponse.responseStream = responseStream
  faasReq.responseBytes = 0
  respCounter = new Stream.Transform()
  respCounter._transform = (chunk, encoding, callback) ->
    debug('response chunk', encoding, chunk)
    faasReq.responseBytes += chunk.length
    callback(null, chunk)
  fwdReq.on 'error', (err) ->
    debug('fwdReq.error', err, proxyResponse)
    if responseStarted
      console.error('duplicate ERROR response attempted for '+url)
      responseStream.emit('error', err)
    else
      responseStarted = true
      proxyResponse.error = err.toString()
      cb_proxyResponse(proxyResponse)
    try
      fwdReq.abort()
      fwdReq.destroy()
      requestStream.unpipe()
      requestStream.on('data', (data) -> debug('fwdReq.error data', data) )
      responseStream.end()
    catch e
      console.error(e)
    Logger.logHttpResponseError(faasReq, proxyResponse)

  fwdReq.on 'response', (response) ->
    Logger.logHttpResponseStart(faasReq, response)
    debug('fwdReq.response', response.headers, response.statusCode, response.statusMessage)
    # debug('fwdReq.response.proxyResponse', proxyResponse)
    if responseStarted
      console.error('duplicate SUCCESS response attempted for '+url)
      return
    responseStarted = true
    proxyResponse.lag = faasReq.ts.responseStart - faasReq.ts.requestStart
    proxyResponse.headers = response.headers
    proxyResponse.statusCode = response.statusCode
    proxyResponse.statusMessage = response.statusMessage
    stream = fwdReq
    if isRewriteHostOn
      # TODO: Rewrite 301/302 redirects?
      ce = proxyResponse.headers['content-encoding']?.toLowerCase()
      isGzip = (ce == 'gzip' || ce == 'x-gzip')
      isDeflate = (ce == 'deflate')
      ct = proxyResponse.headers['content-type']?.toLowerCase() || ''
      canRewrite = (ct.indexOf('text/') > -1 || ct.indexOf('javascript') > -1 || ct.indexOf('js') > -1)
      canRewrite &&= (!ce || ce == 'identity' || isGzip || isDeflate)
      if canRewrite
        delete proxyResponse.headers['content-length']
        proxyResponse.headers['transfer-encoding'] = 'chunked'
        proxyResponse.headers['accept-ranges'] = 'none'
        if isGzip || isDeflate
          stream = fwdReq.pipe(Zlib.createUnzip())
        stream = stream.pipe(newRewriterStream(faasReq.origHost, faasReq.origProto))
        if isGzip
          stream = stream.pipe(Zlib.createGzip())
        else if isDeflate
          stream = stream.pipe(Zlib.createDeflate())
    cb_proxyResponse(proxyResponse)
    stream.pipe(respCounter).pipe(responseStream)

  fwdReq.on 'end', ->
    Logger.logHttpResponseEnd(faasReq)

  requestStream.pipe(reqCounter).pipe(fwdReq)

newRewriterStream = (host, proto) ->
  queue = ''
  if upstreamProto != proto
    hostsReStr = regexEscape(host)
    if Base.env.IX_PROXY_REWRITE_PROTO_HOSTS_REGEX
      hostsReStr += "|"+Base.env.IX_PROXY_REWRITE_PROTO_HOSTS_REGEX
    changeProtoOriginsRe = new RegExp("\\b#{upstreamProto}(#{rewriteProtoDelimReStr})(#{hostsReStr})\\b", "ig")
    debug('changeProtoOriginsRe', changeProtoOriginsRe)
  rewriter = new Stream.Transform()
  rewriter._transform = (chunk, encoding, callback) ->
    debug('rewriter chunk', encoding, chunk)
    queue += chunk.toString('utf8')
    queue = queue.replace(upstreamHostRe, host)
    if upstreamProto != proto
      queue = queue.replace(changeProtoOriginsRe, "#{proto}$1$2")
    if rewriteCustomReplacers
      for replacer in rewriteCustomReplacers
        queue = queue.replace(replacer[0], replacer[1])
    if queue.length > 100
      splitPoint = queue.length - 100
      oldChunk = queue.slice(0, splitPoint)
      queue = queue.slice(splitPoint)
      rewriter.push(new Buffer(oldChunk, 'utf8'))
    callback()
  rewriter._flush = (callback) ->
    debug('rewriter flush')
    rewriter.push(new Buffer(queue, 'utf8'))
    callback()
  rewriter

regexEscape = (str) ->
  # http://stackoverflow.com/questions/2593637/how-to-escape-regular-expression-in-javascript/17326679#17326679
  str.replace(/([.*+?^=!:${}()|\[\]\/\\])/g, "\\$1")
